"""
FiftyOne repository factory.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from fiftyone.factory.repos.delegated_operation_doc import (
    DelegatedOperationDocument,
)
