import o from"./IndexPageQuery.graphql.ddddc526.js";import{r as n,R as e,S as l}from"./index.6894b23b.js";import{N as m,S as c}from"./Nav.57642b04.js";const d=o,u=({prepared:r})=>{var a;const t=n.exports.usePreloadedQuery(d,r),s=(a=t==null?void 0:t.allDatasets)==null?void 0:a.total;return e.createElement(e.Fragment,null,e.createElement(m,{fragment:t,hasDataset:!1}),e.createElement(c,{mode:s===0?"ADD_DATASET":"SELECT_DATASET"}),e.createElement(l,null))};export{u as default};
